from ipaddress import IPv6Address, AddressValueError
from mjail.cmd_helpers import cmd, output, to_tempfile, rc_conf_mod
from mjail.get_jail_conf import get_jail_conf
from mjail.settings import cloned_if
from mjail.pf_conf_split import pf_conf_split
from mjail.jails_network import (
    jails_network4, jails_network6, NoJailNetwork
)
import os
import os.path
import re
from subprocess import CalledProcessError
import warnings

def pf_is_running():
    return output('service', 'pf', 'status').startswith(
        'Status: Enabled'
    )

_ip_reg = r'\d+\.\d+\.\d+\.\d+'

def interface_ip4():
    strg = output('sysrc', '-n', 'ifconfig_%s' % cloned_if()).strip()
    return re.match(
        r'^inet\s(?P<ip4>{ip_reg})\s*(netmask {ip_reg})?$'.format(ip_reg = _ip_reg),
        strg
    ).groupdict()['ip4']

def external_interface_ip6s():
    """
    returns a list of IPv6Address instances: the list of global scope IPv6 addresses
    of the external interface
    """
    result = []
    lines = output('ifconfig', get_ext_if()).split('\n')
    for line in lines:
        mobj = re.match(r'^\s*inet6\s([0-9a-f:]+)', line)
        if mobj:
            try:
                address = IPv6Address(mobj.group(1))
            except AddressValueError:
                pass
            else:
                if address.is_global:
                    result.append(address)
    return result

class NoDefaultGatewayError(Exception):
    pass

def get_ext_if():
    route_table = output('netstat', '-rn')
    try:
        return next(
            line[3]
            for line in (
                l.split() for l in route_table.split('\n')
            )
            if (
                len(line) >= 4
                and
                line[0] == 'default'
                and
                re.match(r'^%s$' % _ip_reg, line[1])
            )
        )
    except StopIteration:
        raise NoDefaultGatewayError

class LocalUnboundManager(object):
    _conf_dir = '/var/mjail/generated_confs/local-unbound/'
    _conf_file = os.path.join(_conf_dir, 'unbound.conf')

    @classmethod
    def enable(cls):
        cmd('mkdir', '-p', cls._conf_dir)
        temp_path = to_tempfile(
            cls._conf(),
            prefix = cls._conf_file
        )
        os.rename(temp_path, cls._conf_file)
        cmd('local-unbound-setup', '-C', cls._conf_dir)

    @classmethod
    def _conf(cls):
        lines = [
            'server:',
            '   access-control: 127.0.0.0/8 allow',
            '   interface: 127.0.0.1'
        ]
        for get_network in (jails_network4, jails_network6):
            try:
                jails_net = get_network()
            except NoJailNetwork:
                pass
            else:
                lines.append(
                    '   access-control: {net} allow'.format(
                        net = str(jails_net)
                    )
                )
                lines.append(
                    '   interface: {ip}'.format(
                        ip = jails_net.network_address + 1
                    )
                )
        return '\n'.join(lines)

class PFManager(object):

    _anchor_conf_file = '/var/mjail/generated_confs/mjail-pf.conf'
    _comment = "# added by mjail, don't modify this line by hand"
    _load_anchor = 'load anchor mjail from "%s" %s' % (_anchor_conf_file, _comment)
    _insert_anchor_filter = "anchor mjail %s" % _comment
    _insert_anchor_nat = "nat-anchor mjail %s" % _comment
    _insert_anchor_rdr = "rdr-anchor mjail %s" % _comment

    @classmethod
    def pf_conf_path(cls):
        try:
            return output('sysrc', '-n', 'pf_rules').strip()
        except CalledProcessError as exc:
            if exc.returncode == 1:
                return '/etc/pf.conf'
            else:
                raise

    @classmethod
    def enable(cls):
        #rc_conf_mod('gateway_enable=YES')
        #rc_conf_mod('net.inet.ip.forwarding=1')
        pf_conf_path = cls.pf_conf_path()
        try:
            pf_conf = [
                line.rstrip()
                for line in open(pf_conf_path).readlines()
                if not line.endswith(cls._comment + "\n")
            ]
        except FileNotFoundError:
            pf_conf = []
        start, translation_rules, filter_rules = pf_conf_split(pf_conf)
        new_conf = (
            start
            +
            [cls._load_anchor, cls._insert_anchor_nat, cls._insert_anchor_rdr]
            +
            translation_rules
            +
            filter_rules
            +
            [cls._insert_anchor_filter]
        )
        new_conf = '\n'.join(new_conf)
        if not new_conf.endswith('\n'): # required by the pf configuration parser
            new_conf += '\n'
        cmd('mkdir', '-p', os.path.dirname(cls._anchor_conf_file))
        if not os.path.exists(cls._anchor_conf_file):
            cls.overwrite_anchor_conf()
        temp_path = to_tempfile(new_conf, prefix = pf_conf_path)
        cmd('pfctl', '-vnf', temp_path) # checking the new conf before replacing the old conf with it
        os.rename(temp_path, pf_conf_path)

        if pf_is_running():
            cmd('pfctl', '-f', pf_conf_path)
        else:
            rc_conf_mod('pf_enable=YES')
            rc_conf_mod('pf_rules=%s' % pf_conf_path)
            cmd('service', 'pf', 'start')

    @classmethod
    def disable(cls):
        pf_conf_path = cls.pf_conf_path()
        try:
            pf_conf = [
                line
                for line in open(pf_conf_path).readlines()
                if not line.endswith(cls._comment + "\n")
            ]
        except FileNotFoundError:
            pass
        else:
            new_conf = ''.join(pf_conf)
            if not new_conf.endswith('\n'):
                new_conf += '\n'
            temp_path = to_tempfile(new_conf, prefix = pf_conf_path)
            cmd('pfctl', '-vnf', temp_path)
            os.rename(temp_path, pf_conf_path)
            if pf_is_running():
                 cmd('pfctl' '-f', pf_conf_path)

    @classmethod
    def _anchor_conf(cls):
        ext_if = get_ext_if()
        filter_rules = []
        translation_rules = []
        jail_conf = get_jail_conf()
        jails = [
            jail_block
            for name, jail_block in jail_conf.jails()
            if (
                jail_block.get('$mjail_managed') == 'yes'
                and (
                    jail_block.get('ip4.addr')
                    or
                    jail_block.get('ip6.addr')
                )
            )
        ]
        cif = cloned_if()
        ext_if_ip6s =  external_interface_ip6s()
        def append_jail_line(rules, jail, *strgs, **formats):
            ips = {
                key: jail[jail_key]
                for key, jail_key in (('ip4', 'ip4.addr'), ('ip6', 'ip6.addr'))
                if jail_key in jail
            }


            for strg in strgs:
                rules.append(
                    strg.format(
                        ext_if = ext_if,
                        cif = cif,
                        **ips,
                        **formats
                    )
                )
        for jail in jails:
            if 'ip4.addr' in jail:
                append_jail_line(
                    translation_rules,
                    jail,
                    'nat on {ext_if} inet from {ip4} to any -> ({ext_if})'
                )
                append_jail_line(
                    filter_rules,
                    jail,
                    'pass quick on {cif} inet proto udp from {ip4} to ({cif}) port 53',
                    'pass quick on {cif} inet proto tcp from {ip4} to ({cif}) port 53'
                )
            if 'ip6.addr' in jail:
                try:
                    ext_if_ip6 = ext_if_ip6s[0]
                except IndexError:
                    raise Exception(
                        "Couldn't find a global scope IPv6 address of the external interface"
                    )
                append_jail_line(
                    translation_rules,
                    jail,
                    'nat on {ext_if} inet6 from {ip6} to any -> %s' % ext_if_ip6
                )
                append_jail_line(
                    filter_rules,
                    jail,
                    'pass quick on {cif} inet6 proto udp from {ip6} to ({cif}) port 53',
                    'pass quick on {cif} inet6 proto tcp from {ip6} to ({cif}) port 53'
                )
            for key in jail:
                if key.startswith('$mjail_rdr_'):
                    proto, host_port = key[len('$mjail_rdr_'):].split('_')
                    assert proto in ('udp', 'tcp')
                    host_port = int(host_port)
                    jail_port = int(jail[key])
                    if 'ip4.addr' in jail:
                        append_jail_line(
                            translation_rules,
                            jail,
                            'rdr pass on {ext_if} inet proto {proto} from any to ({ext_if}) port {host_port} -> {ip4} port {jail_port}',
                            proto = proto,
                            host_port = host_port,
                            jail_port = jail_port,
                        )
                    if 'ip6.addr' in jail:
                        if len(ext_if_ip6s) == 0:
                            raise Exception(
                                "Couldn't find a global scope IPv6 address of the external interface"
                            )
                        if len(ext_if_ip6s) > 1:
                            warnings.warn(
                                "The external interface has many global scope IPv6 addresses. "
                                "The port redirection from port {host_port} to jail {jail_name}:{jail_port} will only occur on "
                                "the address {ext_if_ip6} of the external interface.".format(
                                    host_port = host_port,
                                    jail_name = jail.get('host.hostname', ''),
                                    jail_port = jail_port,
                                    ext_if_ip6 = ext_if_ip6s[0]
                                )
                            )
                        append_jail_line(
                            translation_rules,
                            jail,
                            (
                                'rdr pass on {ext_if} inet6 proto {proto} from any to %s port {host_port} -> {ip6} port {jail_port}'
                                % ext_if_ip6s[0]
                            ),
                            proto = proto,
                            host_port = host_port,
                            jail_port = jail_port,
                        )
        filter_rules.append(
            'pass quick on {cif} from ({cif}) to ({cif}:network)'.format(cif = cif),
            # TODO: allow only the host to access the jails but not the jails between them?
            # (could be an option in `mjail init`)
        )
        ruleset = translation_rules + filter_rules + ['']
        return '\n'.join(ruleset)

    @classmethod
    def overwrite_anchor_conf(cls):
        temp_path = to_tempfile(cls._anchor_conf(), prefix = cls._anchor_conf_file)
        cmd('pfctl', '-vnf', temp_path)
        os.rename(temp_path, cls._anchor_conf_file)

    @classmethod
    def refresh_anchor(cls):
        cls.overwrite_anchor_conf()
        cmd('pfctl', '-a', 'mjail', '-F', 'all', '-f',  cls._anchor_conf_file)



import re
import os
from mjail.cmd_helpers import to_tempfile

class SSHDConf(object):
    def __init__(self, path):
        self._lines = list(map(str.rstrip, open(path).readlines()))
        self._path = path
        
    def set_option(self, option, value):
        rgx = re.compile('^(?P<comment>#?)\s*%s\s+(?P<value>.*)$' % option)
        correct_line = "%s %s" % (option, value)
        def match_line(line):
            mobj = rgx.match(line)
            if mobj:
                return True, correct_line
            else:
                return False, line
        present = False
        for i, line in enumerate(self._lines):
            match, new_line = match_line(line)
            present = present or match
            if match:
                self._lines[i] = new_line
        if not present:
            self._lines.append(correct_line)
            
    def get(self, option, default = None):
        rgx = re.compile('^%s\s+(.*)$' % option)
        for line in self._lines:
            mobj = rgx.match(line)
            if mobj:
                return mobj.group(1)
        return default
            

    def __str__(self):
        return '\n'.join(self._lines)
        
    def overwrite(self):
        temp_path = to_tempfile(str(self), prefix = self._path)
        os.rename(temp_path, self._path)
        
    
    



# stolen from http://stackoverflow.com/a/24176022/1126971

from contextlib import contextmanager
import os

@contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)



import re

is_translation_rule = re.compile(r'^\s*(binat|nat|rdr)\s+').match
is_filter_rule = re.compile(r'^\s*(block|pass)\s+').match

def pf_conf_split(pf_conf_lines):
    # split the conf, expressed as a list of rules, in 3 parts:
    # 1: everything before the translations rules
    # 2: the translations rules
    # 3 the filter rules
    # this function assumes that the conf respects the
    # pf.conf structure described here: https://www.freebsd.org/cgi/man.cgi?query=pf.conf&sektion=5
    step = 1
    part1, part2, part3 = [], [], []
    for line in pf_conf_lines:
        if is_filter_rule(line):
            step = 3
        elif is_translation_rule(line):
            step = 2
        if step == 1:
            part1.append(line)
        elif step == 2:
            part2.append(line)
        else:
            part3.append(line)
    return part1, part2, part3
            



from ipaddress import IPv4Network, IPv4Address, IPv6Network, IPv6Address
import jailconf
from mjail.cd import cd
from mjail.cmd_helpers import cmd, output, rc_conf_mod, to_tempfile
from mjail.get_jail_conf import get_jail_conf
from mjail.settings import cloned_if
from mjail.jails_network import jails_network4, jails_network6, NoJailNetwork
from mjail.sshd_conf import SSHDConf
import os, os.path
import re
import shutil
from subprocess import CalledProcessError
from tempfile import TemporaryDirectory

from mjail.services import PFManager, LocalUnboundManager

class CompatibilityException(Exception):
    pass

def check_compatibility():
    for service in ('local_unbound', 'pf'):
        if not output('sysrc', '-n', '%s_enable' % service).strip() == 'YES':
            raise CompatibilityException(
                "mjail requires {service}\n"
                "run:\n"
                "    sysrc {service}_enable=YES\n"
                .format(service = service)
            )

def init(ip4_network: IPv4Network, ip6_network: IPv6Network):
    if not ip4_network.is_private:
        raise ValueError(
            "The network should be private. "
            "see https://en.wikipedia.org/wiki/Private_network#Private_IPv4_addresses"
        )
    if not ip6_network.is_private or not ip6_network.prefixlen == 64:
        raise ValueError(
            "The network should be a private network as defined in https://tools.ietf.org/html/rfc4193.html"
        )
    check_compatibility()

    rc_conf_mod('cloned_interfaces+=%s' % cloned_if())
    cmd('service', 'netif', 'cloneup')

    address4 = str(ip4_network.network_address + 1)
    netmask = str(ip4_network.netmask)
    rc_conf_mod('ifconfig_%s=inet %s netmask %s' % (cloned_if(), address4, netmask))
    cmd('ifconfig', cloned_if(), 'inet', address4, 'netmask', netmask)

    address6 = str(ip6_network.network_address + 1)
    rc_conf_mod('ifconfig_%s_ipv6=inet6 %s prefixlen 64' % (cloned_if(), address6))
    cmd('ifconfig', cloned_if(), 'inet6', address6, 'prefixlen', '64')

    rc_conf_mod('jail_enable=YES')
    cmd(
        'mkdir', '-p',
        '/var/mjail/instances/',
        '/var/mjail/releases/',
        '/var/mjail/generated_confs/'
    )
    cmd('chmod', '700', '/var/mjail/instances/', '/var/mjail/releases/')
    cmd('chmod', '755', '/var/mjail/', '/var/mjail/generated_confs/')
    try:
        jail_conf = jailconf.load('/etc/jail.conf')
    except FileNotFoundError:
        jail_conf = jailconf.JailConf()
    jail_conf['exec.start'] = '"/bin/sh /etc/rc"'
    jail_conf['exec.stop'] = '"/bin/sh /etc/rc.shutdown"'
    jail_conf['exec.clean'] = True
    jail_conf['mount.devfs'] = True
    jail_conf['path'] = '"/var/mjail/instances/$name"'
    jail_conf.write('/etc/jail.conf')
    release = Release()
    if not release.built():
        release.build()
    LocalUnboundManager.enable()
    PFManager.enable()


class Release(object):

    @classmethod
    def current_release(cls):
        return output('uname', '-r').split('-RELEASE')[0] + '-RELEASE'

    def __init__(self, release = None):
        self._release = release or self.__class__.current_release()

    def __str__(self):
        return self._release

    def _component_url(self, component):
        return (
            'http://ftp.freebsd.org/pub/FreeBSD/releases/amd64/amd64/{release}/{component}'
            .format(
                release = self._release,
                component = component
            )
        )

    @property
    def directory(self):
        return os.path.join('/var/mjail/releases/', self._release)

    def built(self):
        return os.path.exists(os.path.join(self.directory, 'bin/echo'))

    def build(self):
        components = ['base.txz', 'lib32.txz']
        if int(re.match(r'^\d+', self._release).group(0)) < 12: # There's no longer a doc.txz in FreeBSD 12
            components.append('doc.txz')
        cmd('mkdir', '-p', self.directory)
        with TemporaryDirectory() as tempdir:
            with cd(tempdir):
                for component in components:
                    cmd('fetch', self._component_url(component), '-o', component)
                    cmd('tar', 'xvf', component, '-C', self.directory)
                freebsd_update(self.directory, unattended = True)
                cmd(
                    'freebsd-update', '-b', self.directory, 'IDS',
                )

def freebsd_update(directory, unattended):
    if unattended:
        env = os.environ.copy()
        env['PAGER'] = 'cat'
        cmd(
            'freebsd-update', '-b', directory, 'fetch',
            env = env
        )
        try:
            cmd(
                'freebsd-update', '-b', directory, 'install',
                env = env
            )
        except CalledProcessError as exc:
            if exc.returncode == 1: # FreeBSD offers no way to tell if that's because there was no update to install
                pass                # or if it comes from another problem
            else:
                raise
    else:
        cmd('freebsd-update', '-b', directory, 'fetch', 'install')

_ip_reg = r'\d+\.\d+\.\d+\.\d+'

def available_ip4():
    jails_net = jails_network4()
    taken_ip4s = set([jails_net.network_address + 1])
    jail_conf = get_jail_conf()
    for _, jail_block in jail_conf.jails():
        try:
            ip4 = IPv4Address(jail_block['ip4.addr'])
        except KeyError:
            continue
        else:
            taken_ip4s.add(ip4)
    for host in jails_net.hosts():
        if host not in taken_ip4s:
            return host

def available_ip6():
    jails_net = jails_network6()
    taken_ip6s = set([jails_net.network_address + 1])
    jail_conf = get_jail_conf()
    for _, jail_block in jail_conf.jails():
        try:
            ip6 = IPv6Address(jail_block['ip6.addr'])
        except KeyError:
            continue
        else:
            taken_ip6s.add(ip6)
    for host in jails_net.hosts():
        if host not in taken_ip6s:
            return host

class JailAlreadyExists(Exception):
    pass

class IPAlreadyRegistered(Exception):
    pass

class Jail(object):

    @classmethod
    def validate_jail_name(cls, name):
        jailname_regex = re.compile('^(?=.{2})[a-z]{1,16}(\d{0,6})?$')
        # I don't remember why I was so strict about jail names
        if not jailname_regex.match(name):
            raise ValueError(
                "Not a valid jail name. "
                "It should be at least two characters, between 1 and 16 lowercase letters optionnaly followed by at most 6 digits."
            )

    def __init__(self, name):
        self.__class__.validate_jail_name(name)
        self.name = name

    @property
    def directory(self):
        return os.path.join('/var/mjail/instances/', self.name)

    def create(self):
        release = Release()
        if not release.built():
            release.build()

        if os.path.exists(self.directory):
            raise JailAlreadyExists(self.name)

        cmd('cp', '-R', '-v', release.directory, self.directory)

        jail_conf = get_jail_conf()

        if self.name in jail_conf:
            raise JailAlreadyExists(self.name)

        jail_conf[self.name] = jailconf.JailBlock([
            ('$mjail_managed', 'yes'),
            ('$mjail_currently_running_release', str(release)),
            ('host.hostname', self.name)
        ])

        jail_conf.write('/etc/jail.conf')

    def _resolv_conf(self):
        jail_block = get_jail_conf()[self.name]
        lines = []
        for key, get_network in (
            ('ip4.addr', jails_network4),
            ('ip6.addr', jails_network6)
        ):
            if jail_block.get(key):
                try:
                    j_network = get_network()
                except NoJailNetwork:
                    pass
                else:
                    lines.append(
                        "nameserver %s\n" % str(j_network.network_address + 1)
                    )
        return ''.join(lines)

    def _update_resolv_conf(self):
        conf_content = self._resolv_conf()
        with open(os.path.join(self.directory, 'etc', 'resolv.conf'), "w") as fp:
            fp.write(conf_content)

    def minor_upgrade(self, to_version, unattended = False):
        # this function would need to be tested
        freebsd_update_conf = to_tempfile(
            ''.join(
                (re.sub(r'(?<=\b)kernel(?=\b)', '', line) if re.match(r'^Components\s', line) else line)
                for line in
                open('/etc/freebsd-update.conf').readlines()
            )
        )
        try:
            jail_conf = get_jail_conf()
            currently_running = jail_conf[self.name]['$mjail_currently_running_release']
            to_version_major = to_version.split('.')[0]
            running_major = currently_running.split('.')[0]
            if to_version_major != running_major:
                raise Exception(
                    "Can't upgrade from %s to %s. Only minor version upgrade is supported at the moment." % (
                        running_major, to_version_major
                    )
                )
            env = os.environ.copy()
            if unattended:
                env['PAGER'] = 'cat'
            cmd('freebsd-update',
                '-b', self.directory,
                '-f', freebsd_update_conf,
                '-r', to_version, 'upgrade', 'install', '--currently-running', currently_running,
                env = env
            )
            for _ in range(2):
                cmd('freebsd-update',
                    '-b', self.directory,
                    '-f', freebsd_update_conf,
                    'install',
                    env = env
                )
            jail_conf[self.name]['$mjail_currently_running_release'] = to_version
            jail_conf.write('/etc/jail.conf')
        finally:
            os.remove(freebsd_update_conf)

    def start(self):
        cmd('service', 'jail', 'start', self.name)

    def stop(self):
        cmd('service', 'jail', 'stop', self.name)

    def _set_ip(self, ip, version, network):

        if ip not in network:
            raise ValueError(
                "Can't assign this ip to the jail. "
                "The ip should belong to the network %s"
                % network
            )
        jail_conf = get_jail_conf()

        for jail_name, jail_block in jail_conf.jails():
            try:
                ip_addr = jail_block['ip%s.addr' % version]
            except KeyError:
                continue
            if ip_addr == str(ip):
                raise IPAlreadyRegistered
            elif isinstance(ip_addr, (list, tuple)):
                if str(ip) in ip_addr:
                    raise IPAlreadyRegistered

        jail_conf[self.name]['interface'] = cloned_if()
        jail_conf[self.name]['ip%s.addr' % version] = str(ip)

        jail_conf.write('/etc/jail.conf')

        self._update_resolv_conf()

        PFManager.refresh_anchor()

        line = '%s %s\n' % (str(ip), self.name)
        lines = open('/etc/hosts').readlines()
        if line not in lines:
            lines.append(line)
        temp_etc_hosts = to_tempfile(''.join(lines))
        shutil.move(temp_etc_hosts, '/etc/hosts')

    def set_ip4(self, ip4):
        assert isinstance(ip4, IPv4Address)
        self._set_ip(ip4, version = 4, network = jails_network4())

    def set_ip6(self, ip6):
        assert isinstance(ip6, IPv6Address)
        self._set_ip(ip6, version = 6, network = jails_network6())

    def assign_ip4(self):
        self.set_ip4(available_ip4())

    def assign_ip6(self):
        self.set_ip6(available_ip6())

    def delete(self):
        cmd('service', 'jail', 'stop', self.name)
        jail_conf = get_jail_conf()
        try:
            jail_block = jail_conf[self.name]
        except KeyError:
            pass
        else:
            for version in (4, 6):
                try:
                    ip = jail_block['ip%s.addr' % version]
                except KeyError:
                    pass
                else:
                    assert isinstance(ip, str) # list of ips not yet supported by mjail
                    line = '%s %s\n' % (ip, self.name)
                    lines = [l for l in open('/etc/hosts').readlines() if l != line]
                    temp_etc_hosts = to_tempfile(''.join(lines))
                    shutil.move(temp_etc_hosts, '/etc/hosts')
            del jail_conf[self.name]
            jail_conf.write('/etc/jail.conf')

        if os.path.exists(self.directory):
            cmd('chflags', '-R', 'noschg', self.directory)
            cmd('rm', '-rf', self.directory)
        PFManager.refresh_anchor()

    def available_shells(self):
        return list(
            filter(
                os.path.exists,
                (
                    line.strip()
                    for line in (
                        open(os.path.join(self.directory, 'etc/shells'))
                        .readlines()
                    )
                    if line.startswith('/')
                )
            )
        )

    def execute(self, command, *args):
        cmd('jexec', self.name, command, *args)

    def shell(self, shell_path = None):
        shell_path = shell_path or '/bin/csh'
        if shell_path in self.available_shells():
            self.execute(shell_path)
        else:
            raise Exception("No such shell: '%s'." % shell_path)

    def freebsd_update(self, unattended):
        freebsd_update(self.directory, unattended)

    def rdr(self, proto, internet_facing_host_port, jail_port):
        assert proto in ('tcp', 'udp')
        jail_conf = get_jail_conf()
        jail_conf[self.name][
            '$mjail_rdr_%s_%s' % (proto, int(internet_facing_host_port))
        ] = str(int(jail_port))
        jail_conf.write('/etc/jail.conf')
        PFManager.refresh_anchor()

    @classmethod
    def cancel_rdr(self, proto, internet_facing_host_port):
        assert proto in ('tcp', 'udp')
        jail_conf = get_jail_conf()
        for _, jail_block in jail_conf.jails():
            try:
                del jail_block['$mjail_rdr_%s_%s' % (proto, int(internet_facing_host_port))]
            except KeyError:
                pass
        jail_conf.write('/etc/jail.conf')
        PFManager.refresh_anchor()

    def ssh_box(self, public_key, internet_facing_port, jail_port = 22):
        internet_facing_port = int(internet_facing_port)
        host_sshd_conf = SSHDConf('/etc/ssh/sshd_config')
        if internet_facing_port == int(host_sshd_conf.get('Port', '22')):
            raise ValueError("This port is the ssh port used by the host.")
        root_ssh_dir = os.path.join(self.directory, 'root', '.ssh')
        authorized_keys_file = os.path.join(root_ssh_dir, 'authorized_keys')
        os.makedirs(root_ssh_dir, exist_ok = True)
        cmd('chmod', '700', root_ssh_dir)
        with open(authorized_keys_file, "w") as fp:
            fp.write(public_key)
        cmd('chmod', '600', authorized_keys_file)
        s_conf = SSHDConf(os.path.join(self.directory, 'etc/ssh/sshd_config'))
        for option, value in (
            ('PermitRootLogin', 'prohibit-password'),
            ('RSAAuthentication', 'yes'),
            ('PubkeyAuthentication', 'yes'),
            ('PasswordAuthentication', 'no'),
            ('Port', str(jail_port))
        ):
            s_conf.set_option(option, value)
        s_conf.overwrite()
        rc_conf_mod('sshd_enable=YES', f = os.path.join(self.directory, 'etc/rc.conf'))
        self.rdr('tcp', internet_facing_port, jail_port)



from subprocess import check_call, check_output
import os
from tempfile import mkstemp

def cmd(*args, **kwargs):
    return check_call(args, **kwargs)
    
def output(*args, **kwargs):
    return check_output(args, **kwargs).decode('utf-8')
    
def to_tempfile(strg, prefix = None):
    handle, temp_path = mkstemp(prefix = prefix)
    os.write(handle, strg.encode('utf-8'))
    os.close(handle)
    return temp_path
    
def rc_conf_mod(strg, f = '/etc/rc.conf'):
    try:
        rc_conf = open(f).read()
    except FileNotFoundError:
        pass
    else:
        if not rc_conf.endswith('\n'):
            rc_conf += '\n'
            temp_path = to_tempfile(rc_conf, prefix = f)
            os.rename(temp_path, f)
    cmd('sysrc', '-f', f, strg)
    



from ipaddress import IPv4Network, IPv6Network
from mjail.cmd_helpers import output
from mjail.settings import cloned_if
import re
from subprocess import CalledProcessError

_ip_reg = r'\d+\.\d+\.\d+\.\d+'
_ip6_reg = r'[0123456789abcdefABCDEF:]+'

class NoJailNetwork(Exception):
    pass

class NoIPv4JailNetwork(NoJailNetwork):
    pass

class NoIPv6JailNetwork(NoJailNetwork):
    pass

def jails_network4():
    try:
        cloned_if_params = output('sysrc', '-n', 'ifconfig_%s' % cloned_if())
    except CalledProcessError:
        raise NoIPv4JailNetwork
    else:
        gd = (
            re.match(
                r'^inet\s+(?P<inet>{ip_reg})\s+netmask\s+(?P<netmask>{ip_reg})'.format(ip_reg = _ip_reg),
                cloned_if_params
             )
            .groupdict()
        )
        return IPv4Network((gd['inet'], gd['netmask']), strict = False)

def jails_network6():
    try:
        cloned_if_params = output('sysrc', '-n', 'ifconfig_%s_ipv6' % cloned_if())
    except CalledProcessError:
        raise NoIPv6JailNetwork
    else:
        gd = (
            re.match(
                r'^inet6\s+(?P<inet6>{ip6_reg})\s+prefixlen\s+(?P<prefixlen>\d+)'.format(ip6_reg = _ip6_reg),
                cloned_if_params
             )
            .groupdict()
        )
        return IPv6Network(gd['inet6'] + '/' + gd['prefixlen'], strict = False)



import jailconf

def get_jail_conf():
    try:
        return jailconf.load('/etc/jail.conf')
    except FileNotFoundError:
        raise Exception("No /etc/jail.conf. You need to run `mjail init`.")
        




from setuptools import setup

with open('README.rst') as fd:
    long_description = fd.read()

setup(
    name='mjail',
    version='0.3.0',
    install_requires = ['docopt>=0.6.2', 'jailconf>=0.2.1'],
    packages=['mjail'],
    scripts=['bin/mjail'],
    author = 'Benjamin Le Forestier',
    author_email = 'benjamin@leforestier.org',
    url = 'https://github.com/leforestier/mjail',
    keywords = ["jail", "freebsd", "jail.conf"],
    description = "Command line tool to create and manage FreeBSD jails",
    long_description = long_description,
    classifiers = [
        "Intended Audience :: System Administrators",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5', # required for the IpV4Network constructor
        'Programming Language :: Python :: 3.6',
        'Operating System :: POSIX :: BSD :: FreeBSD',
        'Topic :: System :: Operating System',
        'Topic :: System :: Systems Administration'

    ]
)
